@import url('https://fonts.googleapis.com/css2?family=Lato:ital,wght@0,400;0,700;0,900;1,400;1,700;1,900&family=Open+Sans:ital,wght@0,400;0,600;1,400;1,600&display=swap');

/* Adjust logo size in navbar */

.navbar-brand img {
  height: 70px;
}

.navbar-brand {
  height: 70px;
}

/* Slim navbar */

.bd-header .navbar-item {
  max-height: 3rem;
}

/* Allow content to use full width */

.bd-page-width {
  max-width: 100%;
}

.bd-main .bd-content .bd-article-container {
  max-width: fit-content;
}

/* Mobile sidebar as wide as needed for content */

.bd-container .bd-sidebar-primary[open=""] {
  max-width: 100%;
  width: fit-content;
}

/* Desktop sidebar should not take up too much space */

.bd-container .bd-sidebar-primary:not([open=""]) {
  max-width: 35%;
  width: fit-content;
}

/* General styles */

body {
  font-family: 'Open Sans', sans-serif;
  font-size: medium;
}

h1,
h2,
h3,
h4,
h5,
h6 {
  overflow-wrap: break-word;
}

/* Version switcher styles */

.version-switcher__button[data-active-version-name*="devdocs"] {
  background-color: var(--pst-color-warning);
  border-color: var(--pst-color-warning);
  opacity: 0.9;
}

.version-switcher__button:not([data-active-version-name*="stable"]):not([data-active-version-name*="dev"]) {
  background-color: var(--pst-color-danger);
  border-color: var(--pst-color-danger);
  opacity: 0.9;
}

button.btn.version-switcher__button,
button.btn.version-switcher__button:hover {
  color: black;
}
